import numpy as np
import matplotlib.pyplot as plt


def gradient_pbl(
    lidar_profile: np.ndarray,
    min_grad: float = -2,
    max_grad: float = 0.5,
) -> np.ndarray:
    """Gives the pblh heights given profiles

    Args:
        lidar_profile (np.ndarray): 2D array of lidar profile
        max_grad (float, optional): A max thereshold for the gradient avoiding clouds or other interferences. Defaults to None.
        max_height (int, optional): Max height to seek PBL (in meters). Defaults to 3000.

    Returns:
        np.ndarray: 1D array of pbl heights
    """

    safe_profile = lidar_profile.copy()
    safe_profile[safe_profile <= 0] = 1e-10
    dimension = safe_profile.ndim

    if dimension == 1:
        gradient = np.gradient(np.log10(safe_profile))
    else:
        gradient = np.gradient(np.log10(safe_profile))[1]

    # gradient2 = np.gradient(gradient)[1]

    # num = 0
    # final = 300
    # plt.plot(gradient[num][10:final])
    # plt.plot(np.gradient(gradient[num][10:final]))
    # plt.show()

    if max_grad is not None:
        gradient[gradient < min_grad] = 0
    gradient[gradient > 0] = 0
    # gradient2[gradient2 > 0] = 0

    # plt.plot(gradient[num][10:final])
    # plt.plot(gradient2[num][10:final])
    # plt.show()

    if max_grad is not None:
        gradient[gradient > max_grad] = 0

    min_axis = 0 if dimension == 1 else 1
    mins = np.argmin(gradient, axis=min_axis)

    return mins


def variance_pbl(
    lidar_profile: np.ndarray,
    window_size: int = 10,
) -> np.ndarray:

    window_number = lidar_profile.shape[0] // window_size
    window_element = np.arange(window_number) * window_size

    variance = np.zeros([window_number, lidar_profile.shape[1]])

    for i, window in enumerate(window_element):
        start = window
        end = start + window_size

        temp_var = np.var(lidar_profile[start:end, :], axis=0)

        variance[i, :] = temp_var

        # var_window = np.var(lidar_profile[start:end], axis=0)
        # variance = np.hstack([variance, var_window])

    variance_vote = np.argmax(variance, axis=1)

    return window_element, variance_vote


def haar(array_shape: int, a: float = 1, b: float = 1 / 2) -> np.ndarray:
    x = np.arange(0, array_shape)
    return np.piecewise(
        x,
        [
            np.logical_and(b - a / 2 <= x, x <= b),
            np.logical_and(b <= x, x <= b + a / 2),
        ],
        [1, -1, 0],
    )


def wavelet_pbl(
    lidar_profile: np.ndarray,
    a: int = 4,
) -> np.ndarray:
    def single_row_wavelet(row: np.ndarray) -> np.ndarray:
        res = np.zeros(row.shape)

        for [idx], _ in np.ndenumerate(row):
            _fn = row * haar(row.shape[0], a, (2 * idx + 1) / 2)
            _int = np.sum(_fn)
            res[idx] += _int

        res[0:a] = np.nan
        res[-a:] = np.nan

        return res / np.sqrt(a)

    wavelets = np.apply_along_axis(single_row_wavelet, 1, lidar_profile)
    wavelet_vote = np.nanargmax(wavelets, axis=1)

    return wavelet_vote



import typer

from lidar_pbl import LidarDataset

app = typer.Typer()


@app.command()
def quicklook(
    data_dir: str = typer.Argument(..., help="Path to the data directory"),
    dark_current_dir: str = typer.Argument(
        ..., help="Path to the dark current directory"
    ),
    max_height: float = typer.Option(2000, help="Maximum height in the Quicklook"),
    methods: bool = typer.Option(False, help="Plot the methods"),
):
    """
    Command line interface for the lidar_pbl package.
    """
    lidar_dataset = LidarDataset(
        data_dir=data_dir,
        dark_current_dir=dark_current_dir,
    )

    lidar_dataset.quicklook(max_height=max_height)

    if methods:
        lidar_dataset.gradient_pbl(min_height=400, max_height=1250, min_grad=-0.05)
        lidar_dataset.wavelet_pbl(min_height=400, max_height=1250, a_meters=90)
        lidar_dataset.variance_pbl(min_height=400, max_height=1250)

    lidar_dataset.show()


@app.command()
def convert(
    input_file: str = typer.Argument(..., help="Path to the input file"),
    output_file: str = typer.Argument(..., help="Path to the output file"),
):
    """
    Command line interface for the lidar_pbl package.
    """
    typer.echo(f"Input file: {input_file}")
    typer.echo(f"Output file: {output_file}")


def run():
    app()



import numpy as np
import matplotlib.pyplot as plt

from matplotlib import colors
from matplotlib.ticker import FuncFormatter
import pendulum


def plot_profile(
    bin: np.ndarray, bin_res: float = 3.75, max_height: None | float = None
) -> None:
    """
    Plots a raw lidar scan.
    Args:
        bin (np.ndarray): The raw lidar scan.
        bin_res (float, optional): The resolution of the lidar scan. Defaults to 3.75.
    """

    bin_number = np.arange(0, bin.shape[0])
    heights = bin_number * bin_res

    if max_height is None:
        plt.plot(bin, heights)
    else:
        index = np.searchsorted(heights, max_height)
        plt.plot(bin[:index], heights[:index])

    plt.xlabel("Lidar Signal [a.u.]")
    plt.ylabel("Height [m]")


def quicklook(
    bin2d: np.ndarray,
    dates: list[pendulum.DateTime],
    bin_res: float = 3.75,
    max_height: None | float = None,
    bin_zero: int = 0,
):
    """
    Quicklook of the data.
    Args:
        bin2d (np.ndarray): The raw lidar scan.
        bin_res (float, optional): The resolution of the lidar scan. Defaults to 3.75.
        max_height (int | float, optional): The maximum height to plot. Defaults to None.
        bin_zero (int, optional): The number of bins to be removed from the top. Defaults to 0.
    """

    if len(dates) != bin2d.shape[0]:
        raise ValueError("The length of dates does not match the number of rows.")

    bin_number = np.arange(0, bin2d.shape[1])
    heights = bin_number * bin_res

    if max_height is None:
        data, h = (
            bin2d[:, bin_zero:].T,
            heights[bin_zero:],
        )
    else:
        index = np.searchsorted(heights, max_height)
        data, h = bin2d[:, bin_zero:index].T, heights[bin_zero:index]

    params = {
        "aspect": "auto",
        "cmap": "jet",
        "interpolation": "antialiased",
        "origin": "lower",
        "norm": colors.LogNorm(vmax=data.max() * 0.004, clip=True),
    }

    fig, ax = plt.subplots(figsize=(12, 5))

    ax.imshow(data, **params)

    ax.set(xlabel="Time", ylabel="Height [m]", title="Lidar Scan")

    @FuncFormatter
    def format_heights(x, pos=None):
        if x >= h.shape[0]:
            return ""
        return f"{h[int(x)]}"

    @FuncFormatter
    def format_dates(x, pos=None):
        if x >= dates.shape[0]:
            return ""
        return dates[int(x)].format("HH:mm:ss")

    ax.yaxis.set_major_formatter(format_heights)
    ax.xaxis.set_major_formatter(format_dates)
    fig.autofmt_xdate(rotation=45)

    fig.colorbar(ax.images[0], ax=ax, label="Lidar Signal [a.u.]")



import toml
from pathlib import Path


import matplotlib

from .core.lidar_dataset import LidarDataset
from .utils import in_wsl


def get_version() -> str:
    path = Path(__file__).resolve().parents[1] / "pyproject.toml"
    pyproject = toml.loads(open(str(path)).read())
    return pyproject["tool"]["poetry"]["version"]


__version__ = get_version()

if in_wsl():
    matplotlib.use("GTK3Agg")



import pathlib

import numpy as np
from matplotlib import pyplot as plt

from lidar_pbl.core.methods import gradient_pbl, variance_pbl, wavelet_pbl
from lidar_pbl.utils import (
    read_npz,
    plot_profile,
    rcs,
    quicklook,
    read_txts,
)


class LidarDataset:
    def __init__(
        self,
        data_dir: pathlib.Path | str,
        dark_current_dir: pathlib.Path | np.ndarray,
        bin_res: float = 3.75,
        bin_zero: int = 12,
    ):
        """Initialize LidarDataset for data processing

        Args:
            data_path (pathlib.Path | str): The path of the directory if data_type is NPZ, or the path of the txt file if data_type is TXT.
            data_type (InputType, optional): The type of the data. Defaults to InputType.NPZ.
        """
        self._data: np.ndarray = np.empty(0)
        self.bin_res: float = bin_res
        self.bin_zero: int = bin_zero

        if isinstance(data_dir, str | pathlib.Path):
            data_path = pathlib.Path(data_dir)
        else:
            raise TypeError("data_path must be a pathlib.Path or a string")

        if data_path.is_dir():
            cache_file = data_path / ".cache.npz"

            if cache_file.exists():
                temp_data, dates = read_npz(cache_file)
            else:
                temp_data, dates = read_txts(data_path)
        else:
            raise FileNotFoundError("The data directory does not exist")

        if isinstance(dark_current_dir, str | pathlib.Path):
            dark_current_path = pathlib.Path(dark_current_dir)
        else:
            raise TypeError("dark_current_path must be a pathlib.Path or a string")

        if dark_current_path.is_dir():
            dc_cache_file = dark_current_path / ".cache.npz"

            if dc_cache_file.exists():
                dark_current_matrix, _ = read_npz(dark_current_path / ".cache.npz")
            else:
                dark_current_matrix, _ = read_txts(dark_current_path)

        dark_current_mean = dark_current_matrix.mean(axis=0)
        dark_current = np.tile(dark_current_mean, (temp_data.shape[0], 1))

        try:
            self.data = temp_data - dark_current
        except ValueError:
            raise ValueError("The data and dark current must have the same shape")

        self.dark_current = dark_current
        temp_data: np.ndarray = temp_data - self.dark_current
        temp_data[temp_data < 0] = 0
        self.data = temp_data
        self.dates = dates
        self.heights = np.arange(temp_data.shape[1]) * self.bin_res

    @property
    def data(self) -> np.ndarray:
        return self._data

    @data.setter
    def data(self, data):
        self.rcs = rcs(data)
        self._data = data

    def plot_profile(
        self,
        bin_number: int = 0,
        max_height: float = None,
    ):
        """Plot the profile of the data.

        Args:
            bin_number (int, optional): The number of bins. Defaults to 0.
            max_height (int | float, optional): The maximum height to plot. Defaults to None.
        """
        plot_profile(self.rcs[bin_number], max_height=max_height)

    def quicklook(self, max_height: float = None):
        """Quicklook of the data.

        Args:
            bin_number (int, optional): The number of bins. Defaults to 0.
            max_height (int | float, optional): The maximum height to plot. Defaults to None.
        """

        quicklook(
            self.rcs,
            self.dates,
            max_height=max_height,
            bin_zero=self.bin_zero,
            bin_res=self.bin_res,
        )

    def gradient_pbl(
        self, min_height: float = 0, min_grad: float = -0.08, max_height: float = 3000
    ):
        """Gradient PBL height criterias

        Args:
            min_grad (float, optional): _description_. Defaults to 0.08.
            max_height (int, optional): _description_. Defaults to 3000.
        """
        height_index = np.searchsorted(self.heights, max_height)
        min_height_index = np.searchsorted(self.heights, min_height)
        points = gradient_pbl(
            self.rcs[:, min_height_index:height_index], min_grad=min_grad
        )

        plt.scatter(
            np.arange(points.size),
            points - self.bin_zero + min_height_index,
            marker="^",
            label=f"Gradient method",
            alpha=0.5,
            s=15,
        )

    def variance_pbl(self, max_height=3000, min_height=0, window_size=10):
        height_index = np.searchsorted(self.heights, max_height)
        min_height_index = np.searchsorted(self.heights, min_height)
        element, variance = variance_pbl(
            self.rcs[:, min_height_index:height_index], window_size=window_size
        )

        plt.scatter(
            element + window_size // 2,
            variance - self.bin_zero + min_height_index,
            marker="o",
            label=f"Variance method",
            alpha=0.5,
            s=15,
            c="g",
        )

    def wavelet_pbl(self, max_height=3000, min_height=0, a_meters: float = 20):
        height_index = np.searchsorted(self.heights, max_height)
        min_height_index = np.searchsorted(self.heights, min_height)
        a = int(a_meters // self.bin_res)

        points = wavelet_pbl(self.rcs[:, min_height_index:height_index], a=a)

        plt.scatter(
            np.arange(points.size),
            points - self.bin_zero + min_height_index,
            marker="^",
            label=f"Wavelet method",
            alpha=0.5,
            s=15,
            c="k",
        )

    def show(self, legend=True) -> None:
        """Show the data.

        Args:
            bins (int, optional): The number of bins. Defaults to 0.
        """
        if legend:
            plt.legend()
        plt.show()



from pathlib import Path
from platform import uname

import toml
import numpy as np


def in_wsl() -> bool:
    return "microsoft-standard" in uname().release


def moving_average(x: np.ndarray, w: int = 3) -> np.ndarray:
    return np.convolve(x, np.ones(w), "valid") / w



import re
import pathlib

import pendulum
import numpy as np


def read_txts(
    dirpath: pathlib.Path | str,
) -> tuple[np.ndarray, list[pendulum.DateTime]]:
    """Opens all txt files in a directory and return a 2 numpy array of them sorted and a list of dates.

    Args:
        dirpath (pathlib.Path | str): The dir where those txt files are.

    Returns:
        tuple[np.ndarray, list[Any]]: 2D values for time and space dependence and a list of dates.
    """
    txts = list(map(str, pathlib.Path(dirpath).glob("*.txt")))
    txts.sort()
    res = [
        re.match(
            r"^.+(RS|DC)(\d{2})(\d{1})(\d{2})(\d{2})\.(\d{2})(\d{2})\d+\.txt$", txt
        )
        for txt in txts
    ]
    dates = [
        pendulum.parse(
            f"20{res.group(2)}-0{res.group(3)}-{res.group(4)}T{res.group(5)}:{res.group(6)}:{res.group(7)}"
        )
        for res in res
        if res is not None
    ]
    data = [np.loadtxt(txt, skiprows=5) for txt in txts]

    cache_file = pathlib.Path(dirpath) / ".cache.npz"
    np.savez(cache_file, data=data, dates=dates)

    return data, dates


def read_npz(
    filepath: pathlib.Path | str,
) -> tuple[np.ndarray, list[pendulum.DateTime]]:
    """Opens a npz file and return a 2 numpy array of them sorted and a list of dates.

    Args:
        filepath (pathlib.Path | str): The filepath of the npz file.

    Returns:
        tuple[np.ndarray, list[Any]]: 2D values for time and space dependence and a list of dates.
    """
    npz = np.load(filepath, allow_pickle=True)

    return npz["data"], npz["dates"]


def txt_to_npz(
    dirpath: pathlib.Path | str, filepath_output: pathlib.Path | str
) -> None:
    """Converts a lot of txt files to a single npz file.
    Which can be opened with:
    lidar = np.load('filepath_output.npz')
    lidar['data'] is a list of numpy arrays.
    lidar['dates'] is a list of pendulum dates.
    Args:
        dirpath (pathlib.Path | str): _description_
        filepath_output (pathlib.Path | str): _description_

    Returns:
        _type_: None
    """
    txts = list(map(str, pathlib.Path(dirpath).glob("*.txt")))
    txts.sort()
    res = [
        re.match(r"^.+RS/RS(\d{2})(\d)(\d{2})(\d{2})\.(\d{2})(\d{2})\d+\.txt$", txt)
        for txt in txts
    ]
    dates = [
        pendulum.parse(
            f"20{res.group(1)}-0{res.group(2)}-{res.group(3)}T{res.group(4)}:{res.group(5)}:{res.group(6)}"
        )
        for res in res
    ]
    data = [np.loadtxt(txt, skiprows=5) for txt in txts]
    filepath_output = (
        filepath_output
        if filepath_output.endswith(".npz")
        else filepath_output + ".npz"
    )
    np.savez(filepath_output, data=data, dates=dates)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lidar_pbl', 'lidar_pbl.cli', 'lidar_pbl.core', 'lidar_pbl.utils']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.5.1,<4.0.0',
 'numpy>=1.22.3,<2.0.0',
 'pandas>=1.4.1,<2.0.0',
 'pendulum>=2.1.2,<3.0.0',
 'rich>=12.0.0,<13.0.0',
 'toml>=0.10.2,<0.11.0',
 'typer>=0.4.0,<0.5.0']

entry_points = \
{'console_scripts': ['lidar-pbl = lidar_pbl.cli.app:run']}

setup_kwargs = {
    'name': 'lidar-pbl',
    'version': '0.1.1',
    'description': 'Package in development that contains functions and command line utitlities to handle lidar data and PBL height calculations',
    'long_description': None,
    'author': 'Juan Diego',
    'author_email': 'jdidelarc@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jdlar1/lidar-pbl',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import numpy as np


def rcs(
    bins: np.ndarray, bin_res: float = 3.75, background_height: float = 8000
) -> np.ndarray:
    """Calculates the range-corrected signal.
    Args:
        bins (np.ndarray): The raw lidar scan.
        bin_res (float, optional): The resolution of the lidar scan. Defaults to 3.75.
    """

    bin_number = np.arange(0, bins.shape[1])
    heights = bin_number * bin_res
    heights = np.tile(heights, (bins.shape[0], 1))

    return bins * heights * heights
